/***********************************************************************/
/*                                                                     */
/*                         C_data_user_rationnal.c                     */
/*                                                                     */
/*  example of definition of the C type rationnal                      */
/*                                                                     */
/*  authors: D. Gaffe and A. Ressouche                                 */
/*                                                                     */
/*  version: 03/06/2022                                                */
/***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "C_data_user_rationnal.h"

const rationnal zero_rationnal={0,1};

unsigned long long int PGCD(unsigned long long int a, unsigned long long int b)
{
  unsigned long long int r;
  do
  {
    r= a % b;
    a= b;
    b= r;
  } while (r != 0);
  return a;
}

unsigned long long int PPCM(unsigned long long int a, unsigned long long int b)
{
  unsigned long long int resu;
  resu= (a/PGCD(a,b)) * b; /* and not (a*b)/PGCD(a,b) */
  return resu;
}

rationnal normalisation(rationnal x)
{
  rationnal resu;
  unsigned long long int p;
  if(x.num != 0 && x.denom !=0)
  {  
    p= PGCD(llabs(x.num),x.denom);
  }
  else
  {
    p=1; /* no change */
  }
  resu.num= x.num / p;
  resu.denom= x.denom / p;
  return resu;
}


rationnal copy_rationnal(rationnal x)
{
  return x;
}

rationnal add_rationnal(rationnal x, rationnal y)
{
  rationnal resu;
  unsigned long long int p;
  long long int coeffx, coeffy;
  if((x.denom != 0) && (y.denom !=0))
  {
    p= PPCM(x.denom,y.denom);
    coeffx= p / x.denom;
    coeffy= p / y.denom;
    resu.num= x.num * coeffx + y.num * coeffy;
    resu.denom= p;
  }
  else
  {
    resu.num= 0;
    resu.denom= 1;
    fprintf(stderr, "ADDITION ERROR betwenn %lld/%lld and %lld/%lld !\n",
	   x.num,x.denom,y.num,y.denom);
  }
  return normalisation(resu);
}

rationnal sub_rationnal(rationnal x, rationnal y)
{
  rationnal resu;
  unsigned long long int p;
  long long int coeffx, coeffy;
  if((x.denom != 0) && (y.denom !=0))
  {
    p= PPCM(x.denom,y.denom);
    coeffx= p / x.denom;
    coeffy= p / y.denom;
    resu.num= x.num * coeffx - y.num * coeffy;
    resu.denom= p;
  }
  else
  {
    resu.num= 0;
    resu.denom= 1;
    fprintf(stderr, "SUBTRACTION ERROR betwenn %lld/%lld and %lld/%lld !\n",
	   x.num,x.denom,y.num,y.denom);
  }
  return normalisation(resu);
}

rationnal neg_rationnal(rationnal x)
{
  rationnal res;
  res.num  = 0 - x.num;
  res.denom = x.denom;
  return res;
}

rationnal mult_rationnal(rationnal x,rationnal y)
{
  rationnal resu;
  resu.num= x.num * y.num;
  resu.denom= x.denom * y.denom;
  return normalisation(resu);
}

rationnal div_rationnal(rationnal x,rationnal y)
{
  rationnal resu;
  resu.num= x.num * y.denom;
  resu.denom= x.denom * y.num;
  return normalisation(resu);
}

int eq_rationnal(rationnal x,rationnal y)
{
  return (x.num == y.num) && (x.denom == y.denom);
}

int neq_rationnal(rationnal x,rationnal y)
{
  return (x.num != y.num) || (x.denom != y.denom);
}

int less_rationnal(rationnal x,rationnal y)
{
  rationnal resu;
  resu= sub_rationnal(x,y);
  return (resu.num < 0);
}

int lesseq_rationnal(rationnal x,rationnal y)
{
  return less_rationnal(x,y) || eq_rationnal(x,y);
}

int sup_rationnal(rationnal x,rationnal y)
{
  rationnal resu;
  resu= sub_rationnal(x,y);
  return (resu.num > 0);
}

int supeq_rationnal(rationnal x,rationnal y)
{
  return sup_rationnal(x,y) || eq_rationnal(x,y);
}
